from models import basic_user
from creds import db
from datetime import datetime, date
import os
import json


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError ("Type %s not serializable" % type(obj))


def create_dump():
    with db.connect() as conn:
        users = basic_user.select().where(basic_user.c.id > 0)
        users = conn.execute(users).fetchall()
        if users is not None:
            script_dir = os.path.dirname(__file__)
            rel_path = "/user_dump" + (datetime.now().strftime("%Y.%m.%d.%H.%M.%S")) + ".json"
            json_file = os.path.join(script_dir, rel_path)
            f = open(json_file, "x")
            users_list = []
            for item in users:
                users_list.append({'first_name':item.first_name, 'last_name':item.last_name, 'email':item.email, 'user_name':item.username})
            f.write(json.dumps(users_list))
            # print(users)
            print("jsonstring " + json.dumps(users_list))

            print("dump name is " + f.name)

            f.close()

            return "dump successful"

        return "dump unsuccessful"

def retrieve_latest_dump():
    script_dir = os.path.dirname(__file__)
    rel_path = "/"
    json_file = os.path.join(script_dir, rel_path)
    dir_list = os.listdir(json_file)
    print(json_file)
    latest_dump = ""
    for item in dir_list:
        temp_file = os.path.join(json_file, item)
        print(temp_file + " , " +  " , " + str(temp_file.__contains__("user_dump")) + " , " + str((latest_dump < temp_file)))
        if temp_file.__contains__("user_dump") and latest_dump < temp_file:
            latest_dump = temp_file
            print(latest_dump)
    if latest_dump != "":
        jsondata = json.load(open(latest_dump))

        return jsondata
    else:

        return None


from creds import db
from models import song_list, playlist, playlist_content
from song_interactions import find_song_by_id


# Playlist CRUD
def new_playlist(name, user_id):
    # Connect to db
    with db.connect() as conn:
        # Make new playlist record with name and user_id
        new_playlist = playlist.insert().values(playlist_name=name,
                                                user_id=user_id)
        # Execute new record
        conn.execute(new_playlist)

        return f"Made new playlist for user:{user_id}"


def get_user_playlists(user_id):
    with db.connect() as conn:
        user_playlists = playlist.select().where(playlist.c.user_id == user_id)
        return conn.execute(user_playlists).fetchall()


def update_playlist_name(playlist_id, new_name, user_id):
    with db.connect() as conn:
        if user_owns_playlist(playlist_id, user_id):
            update = playlist.update().where(playlist.c.id == playlist_id).values(playlist_name=new_name)
            conn.execute(update)
            return True
        return False


def delete_playlist(playlist_id, user_id):
    with db.connect() as conn:
        if user_owns_playlist(playlist_id, user_id):
            conn.execute(playlist.delete().where(playlist.c.id == playlist_id, playlist.c.user_id == user_id))

            return f"Deleted Playlist @ id:{playlist_id}"

        return f"ERROR: Playlist @ id:{playlist_id} not found or not owned"


def playlist_exists(playlist_id):
    with db.connect() as conn:
        query = playlist.select().where(playlist.c.id == playlist_id)
        return conn.execute(query).fetchone() is not None  # Returns true if playlist is not None


def user_owns_playlist(playlist_id, user_id):
    with db.connect() as conn:
        query = playlist.select().where(playlist.c.id == playlist_id, playlist.c.user_id == user_id)
        return conn.execute(query).fetchone() is not None  # Returns true if playlist is not None


# Playlist content CRUD
def show_playlist_content(playlist_id):
    with db.connect() as conn:
        query = playlist_content.select().where(playlist_content.c.playlist_id == playlist_id)
        result = conn.execute(query).fetchall()
        lst = []
        for song_id in result:
            query = song_list.select().where(song_list.c.genius_id == song_id[1])
            lst.append(conn.execute(query).fetchone())
        return lst


def add_song_to_playlist(song_id, playlist_id, user_id):
    with db.connect() as conn:
        if user_owns_playlist(playlist_id, user_id) and find_song_by_id(song_id):
            try:
                query = playlist_content.insert().values(playlist_id=playlist_id,
                                                         song_id=song_id)
                conn.execute(query)
            except:
                return False
            return True
        return False


def remove_song_from_playlist(song_id, playlist_id, user_id):
    with db.connect() as conn:
        if is_song_in_playlist(song_id, playlist_id) & user_owns_playlist(playlist_id, user_id):
            query = playlist_content.delete().where(playlist_content.c.playlist_id == playlist_id,
                                                    playlist_content.c.song_id == song_id)
            conn.execute(query)

            return f"Delete song @ id:{song_id} from playlist @ id:{playlist_id}"

        return f"Song @ id:{song_id} is not in playlist @ id:{playlist_id} or playlist not owned by user"


def is_song_in_playlist(song_id, playlist_id):
    with db.connect() as conn:
        query = playlist_content.select().where(playlist_content.c.playlist_id == playlist_id,
                                                playlist_content.c.song_id == song_id)
        return conn.execute(query).fetchone() is not None  # Returns true if playlist is not None



from sqlalchemy import exc
from creds import db
from models import basic_user
import find_user as find_user



# Create
def add_user(first_name, last_name, email, username, password):
    with db.connect() as conn:
        try:
            new_user = basic_user.insert().values(first_name=first_name,
                                                  last_name=last_name,
                                                  email=email,
                                                  username=username,
                                                  password=password)
            conn.execute(new_user)

            return "User Added"
        except exc.SQLAlchemyError:

            return "ERROR: " + str(exc.SQLAlchemyError)


# Read
def get_users():
    with db.connect() as conn:
        try:
            select = basic_user.select()
            result = conn.execute(select)

            return result.fetchall()
        except exc.SQLAlchemyError as e:

            return "ERROR: " + str(e)


# Read user @ id
def get_user(user_id):
    with db.connect() as conn:
        try:
            select = basic_user.select().where(basic_user.c.id == user_id)

            return conn.execute(select).fetchone()
        except exc.SQLAlchemyError as e:

            return "ERROR: " + str(e)


# Update first/last @ id
def update_user(user_id, first, last):
    with db.connect() as conn:
        try:
            if find_user.by_id(user_id):
                update = basic_user.update().where(basic_user.c.id == user_id).values(first=first, last=last)
                conn.execute(update)

                return f"User @ id:{user_id} was updated"

            return f"ERROR: User @ id:{user_id} not found"
        except exc.SQLAlchemyError:

            return "ERROR: " + str(exc.SQLAlchemyError)


def delete_user(user_id):
    with db.connect() as conn:
        if find_user.by_id(user_id):
            conn.execute(basic_user.delete().where(basic_user.c.id == user_id))

            return f"Deleted user @ id:{user_id}"

        return f"ERROR: User @ id:{user_id} is not found"



from apscheduler.schedulers.background import BackgroundScheduler
from dump_user import create_dump

scheduler = BackgroundScheduler()

def run_jobs():
    scheduler.start()
    

def add_jobs():
    scheduler.add_job(create_dump, 'interval', minutes=1)


from creds import db
from models import basic_user




def by_id(user_id):
    with db.connect() as conn:
        user = basic_user.select().where(basic_user.c.id == user_id)
        result = conn.execute(user)

        if len(result.fetchall()) == 0:

            return False
        return True


def by_username(username):
    with db.connect() as conn:
        query = basic_user.select().where(basic_user.c.username == username)

        return conn.execute(query).fetchone() is not None  # Returns true if email found


def by_email(email):
    with db.connect() as conn:
        query = basic_user.select().where(basic_user.c.email == email)

        return conn.execute(query).fetchone() is not None  # Returns true if email found


def by_first_and_last(firstname, lastname):
    with db.connect() as conn:
        user = basic_user.select().where(basic_user.c.first_name == firstname & basic_user.c.last_name == lastname)
        result = conn.execute(user)

        if len(result.fetchall()) == 0:

            return False
    return True



from sqlalchemy import exc
from creds import db
from models import basic_user, logged_in_user
from find_user import by_username, by_email
from basic_crud import get_user
import secrets
from datetime import datetime, timedelta


def login(username, password):
    with db.connect() as conn:
        if by_username(username):
            query = basic_user.select().where(basic_user.c.username == username, basic_user.c.password == password)
            result = conn.execute(query).fetchone()
            user_found = result is not None  # Return true if result has content
            if user_found:
                update_time = basic_user.update().where(basic_user.c.username == username, basic_user.c.password ==
                                                        password).values(last_login=datetime.now())
                conn.execute(update_time)
                user_id = basic_user.select().where(basic_user.c.username == username,
                                                    basic_user.c.password == password)
                user_id = conn.execute(user_id).fetchone()[0]

                session_token = user_session_token(user_id)

                return [user_found, session_token]
            return [user_found]  # Sending as array since front_end wont know if it's array or not


def logout(session_token):
    with db.connect() as conn:
        # Write code
        if user_session_valid(session_token):
            query = logged_in_user.delete().where(logged_in_user.c.session_token == session_token)
            conn.execute(query)

            return True
        return False


def register(username, firstname, lastname, email, password):
    with db.connect() as conn:
        if not by_email(email):
            result_data = basic_user.insert().values(first_name=firstname,
                                                     last_name=lastname,
                                                     username=username,
                                                     email=email,
                                                     password=password,
                                                     )
            conn.execute(result_data)

            return True
        return False


def user_session_token(user_id):
    with db.connect() as conn:
        query = logged_in_user.select().where(logged_in_user.c.user_id == user_id)
        if conn.execute(query).fetchone() is None:
            session_token = secrets.token_hex(5)

            add_login = logged_in_user.insert().values(user_id=user_id,
                                                       session_token=session_token,
                                                       token_expiry=datetime.now() + timedelta(days=30))
            conn.execute(add_login)

            return session_token
        return conn.execute(query).fetchone()[1]


def user_session_valid(session_token):
    with db.connect() as conn:
        query = logged_in_user.select().where(logged_in_user.c.session_token == session_token)

        return conn.execute(query).fetchone() is not None


def user_info_from_session_token(session_token):
    with db.connect() as conn:
        query = logged_in_user.select().where(logged_in_user.c.session_token == session_token)
        result = conn.execute(query).fetchone()
        if result is not None:
            user_id = result[0]

            return get_user(user_id)

        return None



from creds import db
from models import liked_songs, song_list, news
from datetime import date
import sqlalchemy as sql
from sqlalchemy import update

def get_news_db():
    with db.connect() as conn:
        results = {}
        result = news.select().where(news.c.id!=0)
        results = conn.execute(result).fetchall()
        return results


from creds import db
from models import liked_songs, song_list, news
from datetime import datetime
import sqlalchemy as sql
from sqlalchemy import update

def has_news():
    with db.connect() as conn:
        result = news.select().where(news.c.id!=0)
        result = conn.execute(result).fetchone()
        return result is not None

def seed_news():
    with db.connect() as conn:
        query = news.insert().values(Title="News", Body="Body", Author="Author", Published=datetime.now())
        conn.execute(query)
        query = news.insert().values(Title="News 2", Body="Body 2", Author="Author 2", Published=datetime.now())
        conn.execute(query)
        query = news.insert().values(Title="News 3", Body="Body 3", Author="Author 3", Published=datetime.now())
        conn.execute(query)
        query = news.insert().values(Title="News 4", Body="Body 4", Author="Author 4", Published=datetime.now())
        conn.execute(query)


import time
import os
from celery import Celery
from celery.utils.log import get_task_logger
from dotenv import load_dotenv
from models import create_all
import basic_crud as basic_crud
import find_user as find_user
import user_interactions as user_interactions
import song_interactions as song_interactions
import playlist_interactions as play_inter
from databaseseed import has_news, seed_news
from socialTasks import get_news_db
from dump_user import create_dump, retrieve_latest_dump
from run_tasks import run_jobs, add_jobs

load_dotenv()

logger = get_task_logger(__name__)

app = Celery('task',
             broker=os.getenv("BROKER_URL"),
             backend='db+postgresql+psycopg2://' + os.getenv("DATABASE_URL"))


# Creates tables
@app.task()
def create_db():
    create_all()
    add_jobs()
    run_jobs()

@app.task()
def get_latest_dump():
    return retrieve_latest_dump()

# User CRUD
# Create
@app.task()
def add_user(first_name, last_name, email, username, password):
    return basic_crud.add_user(first_name, last_name, email, username, password)


# Read
@app.task()
def get_users():
    return basic_crud.get_users()


# Read user @ id
@app.task()
def get_user(user_id):
    return basic_crud.get_user(user_id)


# Update first/last @ id
@app.task()
def update_user(user_id, first, last):
    return basic_crud.update_user(user_id, first, last)


# Delete
@app.task()
def delete_user(user_id):
    return basic_crud.delete_user(user_id)


# Authentication
@app.task()
def login(username, password):
    return user_interactions.login(username, password)


@app.task()
def register(username, first_name, last_name, email, password):
    return user_interactions.register(username, first_name, last_name, email, password)


@app.task()
def logout(session_token):
    return user_interactions.logout(session_token)


@app.task()
def token_valid(session_token):
    return user_interactions.user_session_valid(session_token)


@app.task()
def user_info_from_session_token(session_token):
    return user_interactions.user_info_from_session_token(session_token)


# User Methods
@app.task()
def find_user_by(method, params):
    if type(params) is not list or tuple:
        return "ERROR: params is not a list or tuple"

    if method == "id":
        return find_user.by_id(params[0])
    if method == "username":
        return find_user.by_username(params[0])
    if method == "first&last":
        return find_user.by_first_and_last(params[0], params[1])

    return False


# Like system functions
@app.task()
def get_liked_songs(song_list, user_id):
    return song_interactions.get_liked_songs(song_list, user_id)


@app.task()
def get_liked_song(song_id, user_id):
    return song_interactions.get_liked_song(song_id, user_id)


@app.task()
def like_song(genius_id, user_id):
    return song_interactions.like_song(genius_id, user_id)


@app.task()
def dislike_song(genius_id, user_id):
    return song_interactions.dislike_song(genius_id, user_id)


@app.task()
def add_song(name, artist, genre, genius_id):
    return song_interactions.add_song(name, artist, genre, genius_id)


@app.task()
def find_song(name, artist):
    return song_interactions.find_song(name, artist)

@app.task()
def update_views(genius_id):
    song_interactions.increaseView(genius_id)
    # to-do call increase views

@app.task()
def get_views(genius_id):
    return song_interactions.getView(genius_id)
    # Celery Test Code

@app.task()
def seed_if_empty():
    if not has_news():
        seed_news()

@app.task()
def get_news():
    return get_news_db()

# Playlist functions
@app.task()
def new_playlist(name, token):
    return play_inter.new_playlist(name, user_interactions.user_info_from_session_token(token)[0])


@app.task()
def get_user_playlists(token):
    return play_inter.get_user_playlists(user_interactions.user_info_from_session_token(token)[0])


@app.task()
def update_playlist_name(token, playlist_id, new_name):
    return play_inter.update_playlist_name(playlist_id, new_name, user_interactions.user_info_from_session_token(token)[0])


@app.task()
def delete_playlist(playlist_id, token):
    return play_inter.delete_playlist(playlist_id, user_interactions.user_info_from_session_token(token)[0])


# Playlist content CRUD
@app.task()
def show_playlist_content(playlist_id):
    return play_inter.show_playlist_content(playlist_id)


@app.task()
def add_song_to_playlist(song_id, playlist_id, token):
    return play_inter.add_song_to_playlist(song_id, playlist_id, user_interactions.user_info_from_session_token(token)[0])


@app.task()
def remove_song_from_playlist(song_id, playlist_id, token):
    return play_inter.remove_song_from_playlist(song_id, playlist_id, user_interactions.user_info_from_session_token(token)[0])

@app.task()
def run_dump():
    return create_dump()


# Celery Test Code
@app.task()
def longtime_add(x, y):
    logger.info('Got Request - Starting work ')
    time.sleep(4)
    logger.info('Work Finished ')
    return x + y
# End of Celery Test Code



from creds import db
from models import liked_songs, song_list
import sqlalchemy as sql



# Likes system
def get_liked_songs(songs_list, user_id):
    with db.connect() as conn:
        results = {}
        for song in songs_list:
            result = liked_songs.select().where(liked_songs.c.song_id == song, liked_songs.c.user_id == user_id)
            result = conn.execute(result).fetchone()
            if result is not None:

                results[song] = True
            else:

                results[song] = False
        return results


def get_liked_song(song_id, user_id):
    with db.connect() as conn:
        result = liked_songs.select().where(liked_songs.c.song_id == song_id, liked_songs.c.user_id == user_id)
        result = conn.execute(result).fetchone()

        return result is not None  # Returns True if there is a song


def like_song(song_id, user_id):
    with db.connect() as conn:
        if not get_liked_song(song_id, user_id):
            query = liked_songs.insert().values(song_id=song_id, user_id=user_id)
            conn.execute(query)

            return f"Added {song_id} for {user_id}"

        return f"User already likes song @ id:{song_id}"

def increaseView(genius_id):
    with db.connect() as conn:
        result = song_list.select().where(song_list.c.genius_id == genius_id)
        result = conn.execute(result).fetchone()
        if result is not None and result.views is not None:
            query = song_list.update().where(song_list.c.genius_id == genius_id).values(views = result.views + 1)
        else:
            query = song_list.update().where(song_list.c.genius_id == genius_id).values(views = 1)
        conn.execute(query)

def getView(genius_id):
    with db.connect() as conn:
        result = song_list.select().where(song_list.c.genius_id == genius_id)
        result = conn.execute(result).fetchone()
        if result is not None:
            return result.views
        else:
            return 0

def dislike_song(song_id, user_id):
    with db.connect() as conn:
        if get_liked_song(song_id, user_id):
            conn.execute(liked_songs.delete().where(liked_songs.c.song_id == song_id, liked_songs.c.user_id == user_id))

            return f"Disliked song @ id:{song_id} for user:{user_id}"

        return "This user hasn't liked this song"


# Song interactions
def add_song(name, artist, genre, genius_id):
    with db.connect() as conn:
        insert = song_list.insert().values(
            name=name,
            artist=artist,
            genre=genre,
            genius_id=genius_id
        )

        conn.execute(insert)


def find_song(name, artist):
    with db.connect() as conn:
        select = song_list.select().where(song_list.c.name == name, song_list.c.artist == artist)
        result = conn.execute(select)
        if len(result.fetchall()) > 0:

            return True
        else:

            return False


def find_song_by_id(song_id):
    with db.connect() as conn:
        query = song_list.select().where(song_list.c.genius_id == song_id)
        result = conn.execute(query)
        return len(result.fetchall()) > 0



from creds import db
from sqlalchemy import Table, Column, String, MetaData, Integer, Identity, ForeignKey, DateTime, func

# SQL Expression Language
meta = MetaData(db)
basic_user = Table('users', meta,
                   Column('id', Integer, Identity('user_id_seq', start=1, increment=1), primary_key=True),
                   Column('first_name', String),
                   Column('last_name', String),
                   Column('email', String),
                   Column('username', String),
                   Column('password', String),
                   Column('last_login', DateTime)
                   )

logged_in_user = Table('logged_in_users', meta,
                       Column('user_id', None, ForeignKey('users.id')),
                       Column('session_token', String),
                       Column('token_expiry', DateTime)
                       )

song_list = Table('songs', meta,
                  Column('id', Integer, Identity('song_id_seq', start=1, increment=1), primary_key=True),
                  Column('name', String),
                  Column('artist', String),
                  Column('genre', String),
                  Column('genius_id', Integer, primary_key=True, unique=True),
                  Column('views', Integer)
                  )

liked_songs = Table('liked_songs', meta,
                    Column('id', Integer, Identity('liked_song_id_seq', start=1, increment=1), primary_key=True),
                    Column('song_id', None, ForeignKey('songs.genius_id')),
                    Column('user_id', None, ForeignKey('users.id'))
                    )

playlist = Table('playlists', meta,
                 Column('id', Integer, Identity('playlist_id_seq', start=1, increment=1), primary_key=True),
                 Column('user_id', None, ForeignKey('users.id')),
                 Column('playlist_name', String)
                 )

playlist_content = Table('playlist_content', meta,
                         Column('playlist_id', None, ForeignKey('playlists.id')),
                         Column('song_id', None, ForeignKey('songs.genius_id'))
                         )
news = Table('news_table', meta,
             Column('id', Integer, Identity('news_id_seq', start=1, increment=1), primary_key=True),
             Column('Title', String),
             Column('Body', String),
             Column('Author', String),
             Column('Published', DateTime)
             )

def create_all():
    meta.create_all()
